/*
 *  Copyright 2017 The WebRTC project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

package tee3.webrtc;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.annotation.Nullable;

public class SoftwareVideoDecoderFactory implements VideoDecoderFactory {
  @Nullable
  @Override
  public VideoDecoder createDecoder(String codecType) {
    if (codecType.equalsIgnoreCase("VP8")) {
      return new VP8Decoder();
    }
    if (codecType.equalsIgnoreCase("VP9") && VP9Decoder.nativeIsSupported()) {
      return new VP9Decoder();
    }
    if (codecType.equalsIgnoreCase("H264") && H264Decoder.nativeIsSupported()) {
      return new H264Decoder();
    }
    if (codecType.equalsIgnoreCase("H265") && H265Decoder.nativeIsSupported()) {
      return new H265Decoder();
    }
    return null;
  }

  @Override
  public VideoCodecInfo[] getSupportedCodecs() {
    return supportedCodecs();
  }

  static VideoCodecInfo[] supportedCodecs() {
    List<VideoCodecInfo> codecs = new ArrayList<VideoCodecInfo>();

    codecs.add(new VideoCodecInfo("VP8", new HashMap<>()));
    if (VP9Decoder.nativeIsSupported()) {
      codecs.add(new VideoCodecInfo("VP9", new HashMap<>()));
    }
    if (H264Decoder.nativeIsSupported()) {
      codecs.add(new VideoCodecInfo("H264", new HashMap<>()));
    }
    if (H265Decoder.nativeIsSupported()) {
      codecs.add(new VideoCodecInfo("H265", new HashMap<>()));
    }

    return codecs.toArray(new VideoCodecInfo[codecs.size()]);
  }
}
